import PyTango
import time

class TangoDev(object):
    """
    A generic tango device
    """
    def __init__(self, name):
        self.dev = None
        self._name = name
        self.initDev()
        
    def initDev(self):
        try:
            self.dev = PyTango.DeviceProxy(self._name)
        except:
            print(('Cannot initialize %s device' % self._name))

    def cmd_exec(self, comm, *args, **kwargs):
        if self.dev is None:
            self.initDev()
        if len(args) > 0:
            try:
                self.dev.command_inout(comm, args[0])
            except Exception as error:
                print(('Failed to execute command %s (with arguments) on device %s.' % (comm, self.dev)))
                print((error[1].desc))
        else:
            try:
                self.dev.command_inout(comm)
            except Exception as error:
                print(('Failed to execute command %s on device %s.' % (comm, self.dev)))
                print((error[1].desc))

    def getAttrValue(self, attr):
        if self.dev is None:
            self.initDev()        
        try:
            temp = self.dev.read_attribute(attr)
            if temp.quality != PyTango.AttrQuality.ATTR_INVALID:
                if type(temp.value) == type('abc'):
                    return temp.value
                try:
                    dataSize = len(temp.value)
                    # If excepting, it is real because reals have no size
                    return temp.value[0]
                except:
                    return temp.value
            else:
                return None
        except:
            self.initDev()
            return None

    def setAttrValue(self, attr, value):
        if self.dev is None:
            self.initDev()        
        try:
            temp = self.dev.write_attribute(attr, value)
        except:
            print(('Cannot set %f for device %s' % (value, self._name)))

    def State(self):
        try:
            temp = self.dev.command_inout('State')
            return str(temp).upper()
        except:
            return '####'
        
    def getProperty(self, propertyName):
        try:
            temp = self.dev.get_property(propertyName)
            if propertyName in list(temp.keys()):
                return float(temp[propertyName][0])
            else:
                return None
        except:
            return None
        
class TangoDevAttRW(TangoDev):
    def __init__(self, name, att_name):
        super(self.__class__, self).__init__(name)
        self.att_name = att_name
        self._value = self.value

    @property
    def value(self):
        return self.getAttrValue(self.att_name)

    @value.setter
    def value(self, value):
        self.setAttrValue(self.att_name, value)

class Shutter(TangoDev):
    
    def OPEN(self):
        """
        Opens the shutter
        """
        self.cmd_exec('Open')
    
    def CLOSE(self):
        """
        Closes the shutter
        """
        self.cmd_exec('Close')
        
    def isEnabled(self):
        """
        Check if the opening of the shutter is enabled
        """
        return self.getAttrValue('AbiGroup')
        
class Stopper(TangoDev):
    
    def __init__(self, stateList, cmdPrefix, idxOpenClose):
        name = 'f/access_control/safety'
        super(self.__class__, self).__init__(name)
        self._stateList = stateList
        self._cmdPrefix = cmdPrefix
        self._cmd_open = self._cmdPrefix + 'Open'
        self._cmd_close = self._cmdPrefix + 'Close'
        self._idxOpenClose = idxOpenClose
    
    def OPEN(self):
        if self._idxOpenClose is None:
            self.cmd_exec(self._cmd_open)
        else:
            for idx in self._idxOpenClose:
                self.cmd_exec(self._cmd_open, idx)
                  
    def CLOSE(self):
        if self._idxOpenClose is None:
            self.cmd_exec(self._cmd_close)
        else:
            for idx in self._idxOpenClose:
                self.cmd_exec(self._cmd_close, idx)
    
    def State(self):
        controllo = list()
        status = list()
        for stopper in self._stateList:
            controllo.append(1)
            status.append(self.getAttrValue(stopper))
        
        if None in status:
            return '####'
        if controllo == status:
            return 'OPEN'
        else:
            return 'CLOSED'
        
class OperationMode(TangoDev):
    
    def __init__(self):
        name = 'usa/mps/plcmps'
        super(self.__class__, self).__init__(name)
        self._att = 'OperationModeCode'
        
    def State(self):
        opMode = self.getAttrValue(self._att)
        if opMode == 1:
            return 'LH'
        elif opMode == 2:
            return 'BC01'
        elif opMode == 3:
            return 'DBD'
        elif opMode == 4:
            return 'FEL01'
        elif opMode == 5:
            return 'FEL02'
        else:
            return 'DBD'
        
class Trigger(TangoDev):
    
    def __init__(self, name, attMaster, delayDevName):
        super(self.__class__, self).__init__(name)
        self._attMaster = attMaster
        self._attMasterRef = self._attMaster + 'Ref'
        self.delayDev = TangoDev(delayDevName)
        self._trigger = self.trigger
        self._reference = self.reference
        self._delayOffset = self.delayOffset
    
    @property
    def trigger(self):
        return self.getAttrValue(self._attMaster)
    
    @property
    def reference(self):
        return self.getAttrValue(self._attMasterRef)
    
    @property
    def delayOffset(self):
        return self.delayDev.getProperty('DelayOffset')
    
    def State(self):
        return abs(self.reference - self.trigger) < 2.0    
    
    def RESET(self):
        if self.reference:
            self.setAttrValue(self._attMaster, self.reference)
            
    def SHIFT(self):
        self.RESET()
        time_shift = self.trigger - self.delayOffset * 12.675
        if time_shift:
            self.setAttrValue(self._attMaster, time_shift)
        else:
            print(('Error trying to shift trigger reference for dev %s.' % self._name))    

class BSTController(object):
    
    maxDelay = 20.
    
    def __init__(self):
        """Constructor"""
        # Devices PIL
        self.PILShutter = Shutter('usa/mps/shutter')

        # Beam stopper
        self.LinacStopper = Stopper(['Linac_bst_1'], 'LinacBst', None)
        self.FEL1Stopper = Stopper(['Undulator_bst1_fel1', 'Undulator_bst2_fel1'], 'UndulatorBst', [1, 2])
        self.FEL2Stopper = Stopper(['Undulator_bst1_fel2', 'Undulator_bst2_fel2'], 'UndulatorBst', [3, 4])
        
        # Shutter Padres
        self.FEL1Shutter = Shutter('tango://srv-tango-padres-01:20000/pfe_f01/interlock/sh_pfe_f01.01')
        self.FEL2Shutter = Shutter('tango://srv-tango-padres-01:20000/pfe_f02/interlock/sh_pfe_f02.01')
        
        # Operation mode (DBD - FEL1 - FEL2 or other)
        self.FERMImode = OperationMode()

        # Trigger SEED
        self.SEEDTrigger = Trigger('sl/timing/timing_sl.01', 'SeedMasterDelay', 'f/timing/rttrigger_sl.01')

        # Time to Go
        self.TimeToGo = TangoDevAttRW('f/modulators/modulators', 'TimeString')

        # FermiStatus
        self.fstatus = TangoDev('f/misc/fermistatus')

        #terfermistatus
        self.terafermistatus = TangoDev('srv-tango-tf-01:20000/tf/mover1d/x_sdiag_tf.01')

        #k2 trigger shift
        self.TriggerShift = TangoDevAttRW('f/timing/rttrigger_mod_kg02.01', 'EnableStaticTriggerShift')

    ##############################
    def trigShiftChanged(self,val):
		self.TriggerShift.value=val

    def trigShiftEnable(self):
		print("enable")
		self.TriggerShift.value=True

    def trigShiftDisable(self):
		print("disable")
		self.TriggerShift.value=False
    ### FEL1 stopper & shutter ###
                 
    def openFEL1stopper(self):
        self.FEL1Stopper.OPEN()
        
    def closeFEL1(self):
        self.FEL1Stopper.CLOSE()       
    
    def openFEL1shutter(self):
        t1 = time.time()
        while True:
            if self.FEL1Shutter.isEnabled():
                break
            else:
                t2 = time.time()
                if abs(t2 - t1) >= self.maxDelay:
                    return
            time.sleep(2)
        self.FEL1Shutter.OPEN()
        
    def openFEL1(self):
        self.openFEL1stopper()
        self.openFEL1shutter()
    
    ############################## 
    
    ### FEL2 stopper & shutter ###
    
    def openFEL2stopper(self):
        self.FEL2Stopper.OPEN()     

    def openFEL2shutter(self):
        t1 = time.time()
        while True:
            if self.FEL2Shutter.isEnabled():
                break
            else:
                t2 = time.time()
                if abs(t2 - t1) >= self.maxDelay:
                    return
            time.sleep(2)
        self.FEL2Shutter.OPEN()
    
    def openFEL2(self):
        self.openFEL2stopper()
        self.openFEL2shutter()
        
    def closeFEL2(self):
        self.FEL2Stopper.CLOSE()
		
##############################################
